import { AnalyticsStateMachineOptions } from '../types/AnalyticsStateMachineOptions';
import { StateMachineCallbacks } from '../types/StateMachineCallbacks';

const REBUFFERING_HEARTBEAT_INTERVALS = [3000, 5000, 10000, 30000, 59700];

export abstract class AnalyticsStateMachine {
  protected stateMachine: StateMachine.StateMachine & { [key: string]: any }
  protected onEnterStateTimestamp: number = 0;

  private rebufferingHeartbeatIntervalHandle?: number;
  private currentRebufferingIntervalIndex: number = 0;

  constructor(protected stateMachineCallbacks: StateMachineCallbacks, opts: AnalyticsStateMachineOptions) {
      this.stateMachine = this.createStateMachine(opts);
  }

  public abstract createStateMachine(opts: AnalyticsStateMachineOptions): StateMachine.StateMachine;
  public abstract callEvent(eventType: string, eventObject: any, timestamp: number): void;
  public abstract sourceChange(config: any, timestamp: number, currentTime?: number): void;
  public abstract updateMetadata(metadata?: any): void;

  protected startRebufferingHeartbeatInterval(reset: boolean = true) {
    this.clearRebufferingHeartbeatInterval(reset);
    this.rebufferingHeartbeatIntervalHandle = window.setInterval(() => {
      if (this.stateMachine.current.toLowerCase() !== 'rebuffering') {
        this.clearRebufferingHeartbeatInterval();
        return;
      }
      const timestamp = new Date().getTime();
      const stateDuration = timestamp - this.onEnterStateTimestamp;
      this.stateMachineCallbacks.heartbeat(stateDuration, this.stateMachine.current.toLowerCase(), { buffered: stateDuration });
      this.onEnterStateTimestamp = timestamp;
      this.currentRebufferingIntervalIndex = Math.min(this.currentRebufferingIntervalIndex + 1, REBUFFERING_HEARTBEAT_INTERVALS.length - 1);
      this.startRebufferingHeartbeatInterval(false);
    }, REBUFFERING_HEARTBEAT_INTERVALS[this.currentRebufferingIntervalIndex]);
  }

  protected clearRebufferingHeartbeatInterval(reset: boolean = true) {
    if (reset) {
      this.currentRebufferingIntervalIndex = 0;
    }
    if(this.rebufferingHeartbeatIntervalHandle != null) {
      window.clearInterval(this.rebufferingHeartbeatIntervalHandle);
      this.rebufferingHeartbeatIntervalHandle = undefined;
    }
  }
}
