import { BackendFactory } from '../js/core/BackendFactory';
import { AnalyticsConfig } from '../js/types/AnalyticsConfig';
import { LicensingResult, LicensingResponse } from '../js/types/LicensingRequest';
import { DebuggingBackend } from '../js/core/DebuggingBackend';
import { SequenceNumberBackend } from '../js/core/SequenceNumberBackend';
import { DisableBackend } from '../js/core/DisableBackend';

import { RemoteBackend } from '../js/core/RemoteBackend';
import { ANALYTICS_BACKEND_BASE_URL } from '../js/utils/Settings';
jest.mock('../js/core/RemoteBackend');

import { LicenseCall } from '../js/utils/LicenseCall';
import { MockDeferredLicenseLoadingAdapterWithSupport } from './utils/MockDeferredLicenseLoadingAdapterWithSupport';
import { LicenseCheckingBackend } from '../js/core/LicenseCheckingBackend';

jest.mock('../js/utils/LicenseCall');
jest.mock('../js/core/DisableBackend');

const granted: LicensingResponse = {
  status: LicensingResult.Granted,
  message: 'here you go',
};

const validLicensingRequest = {
  domain: 'foo.com',
  key: '81354CDA-0EE6-4EDB-9A12-9EC7F80BBFE8',
  version: `1.2.3`,
};

const adapter = new MockDeferredLicenseLoadingAdapterWithSupport(true);

beforeEach(() => {
  (RemoteBackend as any).mockClear();
  (DisableBackend as any).mockClear();
});

describe('BackendFactory', () => {

  beforeAll(() => {
    (LicenseCall as any).mockResolvedValue(granted);
  });

  test('it creates a RemoteBackend and makes a LicenseCall', async () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
    };

    await new BackendFactory().createBackend(config, validLicensingRequest, adapter);

    expect(RemoteBackend).toHaveBeenCalledWith(true, ANALYTICS_BACKEND_BASE_URL, '81354CDA-0EE6-4EDB-9A12-9EC7F80BBFE8');
    expect(LicenseCall).toHaveBeenCalled();
  });

  test('it re-uses existing DisableBackend', async () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      config: {
        backendUrl: 'http://my-backend.com',
      },
    };

    const bf = new BackendFactory();

    const backend1 = bf.createBackend(config, validLicensingRequest, adapter);
    const backend2 = bf.createBackend(config, validLicensingRequest, adapter);

    expect(DisableBackend).toHaveBeenCalledTimes(1);
  });

  test('it creates RemoteBackend with backendUrl from analyticsConfig', async () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      config: {
        backendUrl: 'http://my-backend.com',
      },
    };

    await new BackendFactory().createBackend(config, validLicensingRequest, adapter);

    expect(RemoteBackend).toHaveBeenCalledWith(true, 'http://my-backend.com', '81354CDA-0EE6-4EDB-9A12-9EC7F80BBFE8');
  });

  test('it is of type DebuggingBackend when setting debug = true in the config', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      debug: true,
    };

    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter);

    expect(backend).toBeInstanceOf(DebuggingBackend);
  });

  test('it is of type SequenceNumberBackend', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
    };

    const backend = new BackendFactory().createBackend(config, validLicensingRequest, adapter);
    expect(backend).toBeInstanceOf(SequenceNumberBackend);
  });

  test('it is of type NoOpBackend when setting enabled = false in the config', () => {
    const config: AnalyticsConfig = {
      key: 'someKey',
      config: { enabled: false },
    };

    const backend  = new BackendFactory().createBackend(config, validLicensingRequest, adapter) as any;

    expect(backend.underlying).toBeInstanceOf(DisableBackend);
    const db = backend.underlying as DisableBackend;
    expect(db.enabled).toEqual(false);
  });

});
