import { DisableBackend } from '../js/core/DisableBackend';
import { NoOpBackend } from '../js/core/NoOpBackend';
jest.mock('../js/core/RemoteBackend');

beforeEach(() => {
});

const sample = { platform: 'ios', pageLoadTime: 0, pageLoadType: 0, playerStartupTime: 0 };

describe('DisableBackend', () => {
  describe('sendRequest', () => {
    describe('when enabled', () => {
      test('it calls underlying backend ', () => {
        const remoteBackend = new NoOpBackend();
        remoteBackend.sendRequest = jest.fn();

        const b = new DisableBackend(remoteBackend);
        b.sendRequest(sample);
        expect(remoteBackend.sendRequest).toHaveBeenCalled();
      });
    });

    describe('when disabled', () => {
      const remoteBackend = new NoOpBackend();
      remoteBackend.sendRequest = jest.fn();
      const b = new DisableBackend(remoteBackend);
      b.enabled = false;

      test('it doesnt call underlying backend ', () => {
        b.sendRequest(sample);
        expect(remoteBackend.sendRequest).not.toHaveBeenCalled();
      });
    });

    describe('when first enabled', () => {
      describe('with prior sample with playerStartupTime', () => {
        it('does nothing to either sample', () => {
          const remoteBackend = new NoOpBackend();
          remoteBackend.sendRequest = jest.fn();
          const b = new DisableBackend(remoteBackend);

          b.sendRequest({ ...sample, playerStartupTime: 200 });
          expect(remoteBackend.sendRequest).toHaveBeenCalledWith(expect.objectContaining({
            playerStartupTime: 200,
          }));

          b.sendRequest({ ...sample, videoStartupTime: 100, startupTime: 300 });
          expect(remoteBackend.sendRequest).toHaveBeenCalledWith(expect.objectContaining({
            playerStartupTime: 0,
            videoStartupTime: 100,
            startupTime: 300,
          }));
        });
      });
    });

    describe('when first disabled then enabled', () => {
      describe('without prior sample with playerStartupTime', () => {
        it ('does not set startupTime', () => {
          const remoteBackend = new NoOpBackend();
          remoteBackend.sendRequest = jest.fn();
          const b = new DisableBackend(remoteBackend);

          b.enabled = false;
          b.sendRequest({ ...sample });
          expect(remoteBackend.sendRequest).not.toHaveBeenCalled();

          b.enabled = true;
          b.sendRequest({ ...sample, videoStartupTime: 100});
          expect(remoteBackend.sendRequest).toHaveBeenCalledWith(
            expect.objectContaining({ playerStartupTime: 0, videoStartupTime: 100 }));
        });

      });
      describe('with sample with playerStartupTime', () => {
        it('does nothing to second sample if videoStartupTime is not set', () => {
          const remoteBackend = new NoOpBackend();
          remoteBackend.sendRequest = jest.fn();
          const b = new DisableBackend(remoteBackend);

          b.enabled = false;
          b.sendRequest({ ...sample, playerStartupTime: 200 });
          expect(remoteBackend.sendRequest).not.toHaveBeenCalled();

          b.enabled = true;
          b.sendRequest({ ...sample});
          expect(remoteBackend.sendRequest).toHaveBeenCalledWith(sample);
        });
        describe('stores the playerStartupTime for subsequent sample', () => {
          const remoteBackend = new NoOpBackend();
          remoteBackend.sendRequest = jest.fn();
          const b = new DisableBackend(remoteBackend);

          b.enabled = false;
          b.sendRequest({ ...sample, playerStartupTime: 200 });
          expect(remoteBackend.sendRequest).not.toHaveBeenCalled();

          b.enabled = true;
          b.sendRequest({ ...sample, videoStartupTime: 100 });

          it('sums up video and player to startupTime', () => {
            expect(remoteBackend.sendRequest)
              .toHaveBeenCalledWith(expect.objectContaining({ startupTime: 300 }));
          });
          it('sets correct videoStartupTime', () => {
            expect(remoteBackend.sendRequest)
              .toHaveBeenCalledWith(expect.objectContaining({ videoStartupTime: 100 }));
          });
          it('sets correct playerStartupTime', () => {
            expect(remoteBackend.sendRequest)
              .toHaveBeenCalledWith(expect.objectContaining({ playerStartupTime: 200 }));
          });

          describe('third session with videoStartupTime set', () => {
            it ('should not have playerStartupTime set from previous samples', () => {
              b.sendRequest({ ...sample, videoStartupTime: 400 });
              expect(remoteBackend.sendRequest)
                .toHaveBeenCalledWith(expect.objectContaining({ playerStartupTime: 0 }));
            });
          });
        });
      });
    });
  });
});
