import { AnalyticsConfig } from '../../js/types/AnalyticsConfig';
import { SessionPersistenceHandler } from '../../js/core/SessionPersistenceHandler';

const clearCookies = () => {
  document.cookie = `bitmovin_analytics_uuid=someId; expires=1 Jan 1970 00:00:00 GMT;`;
}

describe('SessionPersistenceHandler', () => {
  describe('Handles analytics configuration', () => {
    test('sets cookiesEnabled to true if no configuration exists', () => {
      const config: AnalyticsConfig = {};

      const persistenceHandler = new SessionPersistenceHandler(config);
      expect(persistenceHandler.cookiesEnabled).toBeTruthy();
    });
    test('sets cookiesEnabled to true if configuration exists with cookies enabled', () => {
      const config: AnalyticsConfig = {
        config: {
          cookiesEnabled: true
        }
      };

      const persistenceHandler = new SessionPersistenceHandler(config);
      expect(persistenceHandler.cookiesEnabled).toBeTruthy();
    });
    test('sets cookiesEnabled to false if configuration exists with cookies disabled', () => {
      const config: AnalyticsConfig = {
        config: {
          cookiesEnabled: false
        }
      };

      const persistenceHandler = new SessionPersistenceHandler(config);
      expect(persistenceHandler.cookiesEnabled).toBeFalsy();
    });
  });

  describe('Cookies enabled', () => {
    test('Generates new UserId if none exists and stores in cookie', () => {
      clearCookies();
      const config: AnalyticsConfig = {
        config: {
          cookiesEnabled: true
        }
      };

      const persistenceHandler = new SessionPersistenceHandler(config);
      const userId = persistenceHandler.userId;

      const cookie = document.cookie;

      expect(userId).not.toBeNull();
      expect(userId.length).toBeGreaterThan(0);
      expect(cookie).toContain(userId);
    });
    test('Takes existing userId from cookie', () => {
      const config: AnalyticsConfig = {};
      document.cookie = 'bitmovin_analytics_uuid=someId';

      const persistenceHandler = new SessionPersistenceHandler(config);
      const userId = persistenceHandler.userId;

      expect(userId).toBe(userId);
    });
  });
  describe('Cookies disabled', () => {
    test('Generates new UserId and does not store in cookie', () => {
      clearCookies();
      const config: AnalyticsConfig = {
        config: {
          cookiesEnabled: false
        }
      };

      const persistenceHandler = new SessionPersistenceHandler(config);
      const userId = persistenceHandler.userId;

      expect(userId).not.toBeNull();
      expect(userId.length).toBeGreaterThan(0);
      expect(document.cookie).not.toContain(userId);
    });
  });
});
